import os
import uuid
from urllib.request import Request, urlopen

import pytest
from boto3.session import boto3
from moto.server import ThreadedMotoServer

from .cloud import S3


@pytest.fixture(scope="session")
def s3_server():
    server = ThreadedMotoServer("127.0.0.1", 0, verbose=False)
    server.start()
    try:
        yield server
    finally:
        server.stop()


@pytest.fixture(scope="session")
def s3_config(s3_server) -> dict:
    host, port = s3_server.get_host_and_port()
    return {
        "endpoint_url": f"http://{host}:{port}",
        "aws_access_key_id": "testing",
        "aws_secret_access_key": "testing",
    }


@pytest.fixture
def s3_client(s3_config):
    return boto3.client("s3", **s3_config)


@pytest.fixture(autouse=True)
def reset_s3_fixture(s3_config):
    req = Request(f"{s3_config['endpoint_url']}/moto-api/reset", method="POST")
    try:
        yield
    finally:
        urlopen(req)


@pytest.fixture
def s3_bucket(s3_client) -> str:
    bucket = "test-bucket"
    s3_client.create_bucket(Bucket=bucket)
    return bucket


@pytest.fixture
def s3_versioned_bucket(s3_client) -> str:
    bucket = "test-bucket-versioned"
    s3_client.create_bucket(Bucket=bucket)
    s3_client.put_bucket_versioning(
        Bucket=bucket, VersioningConfiguration={"Status": "Enabled"}
    )
    return bucket


@pytest.fixture
def make_s3(request):
    def _make_s3():
        if bucket := os.environ.get("DVC_TEST_AWS_REPO_BUCKET"):
            return S3(f"s3://{bucket}/dvc_test_caches/{uuid.uuid4()}")
        config = request.getfixturevalue("s3_config")
        bucket = request.getfixturevalue("s3_bucket")
        return S3(f"s3://{bucket}", config=config)

    return _make_s3


@pytest.fixture
def make_s3_version_aware(request):
    def _make_s3():
        if bucket := os.environ.get("DVC_TEST_AWS_REPO_BUCKET_VERSIONED"):
            return S3(f"s3://{bucket}/dvc_test_caches/{uuid.uuid4()}")
        config = request.getfixturevalue("s3_config")
        bucket = request.getfixturevalue("s3_versioned_bucket")
        return S3(f"s3://{bucket}", config=config)

    return _make_s3


@pytest.fixture
def s3(make_s3):
    return make_s3()


@pytest.fixture
def cloud(make_cloud):
    return make_cloud(typ="s3")


@pytest.fixture
def remote(make_remote):
    return make_remote(name="upstream", typ="s3")


@pytest.fixture
def remote_version_aware(make_remote_version_aware):
    return make_remote_version_aware(name="upstream", typ="s3")


@pytest.fixture
def remote_worktree(make_remote_worktree):
    return make_remote_worktree(name="upstream", typ="s3")


@pytest.fixture
def workspace(make_workspace):
    return make_workspace(name="workspace", typ="s3")



import locale
from functools import cached_property

from dvc.testing.cloud import Cloud
from dvc.testing.path_info import CloudURLInfo


class S3(Cloud, CloudURLInfo):
    def __init__(self, url: str, config=None):
        super().__init__(url)
        self._config = config or {}

    def __truediv__(self, key):
        ret = super().__truediv__(key)
        ret._config = self._config
        ret._s3 = self._s3
        return ret

    @property
    def config(self):
        ret = {"url": str(self)}
        config_map = {
            "access_key_id": "aws_access_key_id",
            "secret_access_key": "aws_secret_access_key",
            "endpointurl": "endpoint_url",
        }
        for key, config_key in config_map.items():
            value = self._config.get(config_key)
            if value is not None:
                ret[key] = value
        return ret

    @cached_property
    def _s3(self):
        import boto3

        return boto3.client("s3", **self._config)

    def is_file(self):
        from botocore.exceptions import ClientError

        if self.path.endswith("/"):
            return False

        try:
            self._s3.head_object(Bucket=self.bucket, Key=self.path)
        except ClientError as exc:
            if exc.response["Error"]["Code"] != "404":
                raise
            return False

        return True

    def is_dir(self):
        path = (self / "").path
        resp = self._s3.list_objects(Bucket=self.bucket, Prefix=path)
        return bool(resp.get("Contents"))

    def exists(self):
        return self.is_file() or self.is_dir()

    def mkdir(self, mode=0o777, parents=False, exist_ok=False):
        assert mode == 0o777
        assert parents

    def write_bytes(self, contents):
        self._s3.put_object(Bucket=self.bucket, Key=self.path, Body=contents)

    def unlink(self, missing_ok: bool = False) -> None:
        if not self.exists():
            if not missing_ok:
                raise FileNotFoundError(str(self))
            return
        self._s3.delete_object(Bucket=self.bucket, Key=self.path)

    def rmdir(self, recursive: bool = True) -> None:
        if not self.is_dir():
            raise NotADirectoryError(str(self))

        path = (self / "").path
        resp = self._s3.list_objects(Bucket=self.bucket, Prefix=path)
        entries = resp.get("Contents")
        if not entries:
            return

        if not recursive:
            raise OSError(f"Not recursive and directory not empty: {self}")

        self._s3.delete_objects(
            Bucket=self.bucket, Delete={"Objects": [{"Key": e["Key"]} for e in entries]}
        )

    def read_bytes(self):
        data = self._s3.get_object(Bucket=self.bucket, Key=self.path)
        return data["Body"].read()

    def read_text(self, encoding=None, errors=None):
        if not encoding:
            encoding = locale.getpreferredencoding(False)
        assert errors is None
        return self.read_bytes().decode(encoding)

    @property
    def fs_path(self):
        return self.bucket + "/" + self.path.lstrip("/")
